//
// See the file LICENSE for redistribution information.
//
// Copyright (c) 2002-2004
//	Sleepycat Software.  All rights reserved.
//
// $Id: LoadContainer.cpp,v 1.16 2004/12/21 16:53:00 gmf Exp $
//
#include "dbxml/DbXml.hpp"

#include <string>
#include <vector>
#include <fstream>

#include "LoadContainer.hpp"

LoadContainer::LoadContainer(const std::string &dBEnvironment, 
			     const std::string &containerName, 
			     LoadContainer::StorageType type) {
	DbEnv *env = new DbEnv(0);
	env->set_cachesize(0, 64*1024, 1);
	env->open(dBEnvironment.c_str(), DB_INIT_MPOOL|DB_CREATE, 0);
	mgr = new DbXml::XmlManager(env, DbXml::DBXML_ADOPT_DBENV);

	DbXml::XmlContainer::ContainerType dbxmlType =
		DbXml::XmlContainer::NodeContainer;
	if(type == LoadContainer::WHOLE_DOCUMENT) {
		dbxmlType = DbXml::XmlContainer::WholedocContainer;
	}
	mgr->setDefaultContainerType(dbxmlType);
	
	conName = containerName;
}

LoadContainer::~LoadContainer() {
	delete mgr;
}

std::string LoadContainer::readFile(const std::string &theFile) const {

	std::string xmlString;

	/* Open as binary to preserve whitespace */
	std::ifstream inFile(theFile.c_str(), std::ios::in | std::ios::binary);
	if(!inFile) {
		std::cerr << "Could not open file '" << 
			theFile << "'. Giving up." << std::endl;
		throw std::exception();
	}

	while(!inFile.eof()) {
		std::string stringBuf;
		std::getline( inFile, stringBuf );
		// use \xA for the eol given that these are XML files.
		stringBuf += '\xA';
		xmlString += stringBuf;
	}

	inFile.close();
	return xmlString;
}


std::string LoadContainer::getNameFromPath(const std::string &theFile) const {
	std::string tmpName = theFile;
	std::string theName;
	int pos = tmpName.rfind("/");
	if(pos == std::string::npos ) {
		pos = tmpName.rfind("\\");
	}
	if(pos == std::string::npos) {
		theName = tmpName;
	} else {
		theName.assign( tmpName, pos+1, tmpName.length() );
	}

	return theName;
}

int LoadContainer::addFiles(const std::vector<std::string > &fileList, 
			    const std::string &pathPrepend) const
{
	try {
		DbXml::XmlContainer container =
			mgr->openContainer(conName, DB_CREATE);
	
		std::cout << "Created the container \"" <<
			container.getName() << "\""
			  << " (" 
			  << (container.getContainerType() ==
			      DbXml::XmlContainer::WholedocContainer ?
			      "whole document" : "node level") 
			  << " storage)" << std::endl;
		
		// Get an update context. Helpful for performance.
		DbXml::XmlUpdateContext updateContext =
			mgr->createUpdateContext();
		
		unsigned int count = 0;
		std::vector<std::string >::const_iterator theFile, 
			end = fileList.end();
		for (theFile = fileList.begin(); theFile != end; ++theFile) {
			std::string totalPath = pathPrepend + *theFile;
			std::string xmlString = readFile(totalPath);
			DbXml::XmlDocument xmlDoc = mgr->createDocument();
			xmlDoc.setContent(xmlString);
			xmlDoc.setName(getNameFromPath(*theFile));
			
			container.putDocument(xmlDoc, updateContext,
					      DbXml::DBXML_GEN_NAME);
			
			std::cout << "\tPut document \"" << xmlDoc.getName()
				  << "\"" << std::endl;
			count++;
		}
		
		std::cout << "\nPut " << count << " documents" << std::endl;
	}
	catch (DbXml::XmlException &e) {
		std::cerr << "Error adding XML data to container: " << 
			conName << std::endl;
		std::cerr << e.what() << std::endl;
		return -1;
	}
	return 0;
}

int LoadContainer::addFileList(const std::string &pathPrepend,
			       const std::string &fileName) const {
	std::ifstream inFile(fileName.c_str() ,std::ios::in);
	if(!inFile) {
		std::cerr << "Could not open file " << fileName 
			  << ". Giving up." << std::endl;
		return -1;
	}

	std::vector<std::string > fileList;
	while(!inFile.eof()) {
		std::string stringBuf;
		std::getline( inFile, stringBuf );
		if(!stringBuf.empty()) {
			fileList.push_back(stringBuf);
		}
	}
	return addFiles(fileList, pathPrepend);
}

void usage()
{
        std::cerr << "This program loads XML data into an identified container and environment.\n"
                  << "The container storage type is specified.\n"
                  << "Provide the directory where you want to place your database environment, the\n"
                  << "name of the container to use, and the xml files you want inserted into the\n"
                  << "container.\n\n"
		
                  << "\t-h - this message\n"
                  << "\t-d - database environment directory \n"
                  << "\t-c - container name\n"
                  << "\t-s - storage type [\"wholedoc\" (*) | \"node\"]\n"
                  << "\t-f - filelist\n"
                  << "\t-p - path to prepend to each of the items in filelist\n"
                  << "file1.xml file2.xml file3.xml .... " << std::endl;
        exit(-1);
}


int main(int argc, char **argv)
{
	int ret = 0;
	std::string path2DbEnv;
	std::string theContainer;
	std::string storageType("wholedoc");
	std::string filePath;
	std::string fileList;
	std::vector<std::string> files2add;
	
	for ( int i=1; i<argc; i++ ) {
		if ( argv[i][0] == '-' ) {
			switch(argv[i][1]) {
			case 'd':
				path2DbEnv = argv[++i];
				break;
			case 'c':
				theContainer = argv[++i];
				break;
			case 's':
				storageType = argv[++i];
				break;
			case 'p':
			{
				// ensure we have a trailing slash
				std::string tmp(argv[++i]);
				int pos = (int) tmp.rfind("/");
				if(pos != tmp.length()-1) {
					pos = (int) tmp.rfind("\\");  // maybe under windows?
				}
				if (pos != tmp.length()-1) {
					tmp.append("/");
				}
				filePath.assign(tmp);
			}
			break;
			case 'f':
				fileList = argv[++i];
				break;
			default:
				usage();
			}
			
		} else {
			files2add.push_back( argv[i] );
		}
	}
	
	//You must provide a container name, a path to a database environment,
	if (path2DbEnv.empty() || theContainer.empty()) {
		usage();
	}
	
	LoadContainer::StorageType type = LoadContainer::WHOLE_DOCUMENT;
	if(!storageType.compare("node")) {
		type = LoadContainer::NODE;
	} else if(storageType.compare("wholedoc")) {
		usage();
	}
	
	LoadContainer lc(path2DbEnv, theContainer, type);
	
	if(!fileList.empty()) {
		ret = lc.addFileList(filePath, fileList);
	}
	if ((ret == 0) && files2add.size()) {
		ret = lc.addFiles(files2add, filePath);
	}
	return ret;
}
    
